# mwprop.ne2001p v1.0 Jan 2024

from numpy import  log10

def calc_mwpsr_taudm(DM, A=2.98e-7, B=3.55e-5, alph=1.4, gam=3.1, sigma=0.76, sw=True):

    """
    Evaluates the fit to pulsar scattering

    tau(1 GHz, ms) = A \times DM^\alpha (1 + B \times DM^\beta) \times 10^{\pm sigma}

    Input: 
        DM = dispersion measure (pc/cc)
        A,B,alph,gam,sigma = values of fitted parameters
        sw = True  ==> use model as is = appropriate for Galactic objects
             False ==> multiply by xsw2pw = 3 = appropriate for extragalactic objects

    Output:
        taudm  = scattering time at 1 GHz in ms
        taudm_minus,plus = scattering time at \mp 1 sigma in log10
        taudm_outer_extend_minus,plus = extension of the low DM part

    """

    # xsw2pw = correction from spherical-wave scattering to plane-wave scattering
    xsw2pw = 3

    taudm = A * DM**alph * (1. + B * DM**gam)
    taudm_minus = 10.**(log10(taudm) - sigma)
    taudm_plus = 10.**(log10(taudm) + sigma)

    tau_outer_extend_minus= 10.**(log10(A*DM**alph) - sigma)
    tau_outer_extend_plus = 10.**(log10(A*DM**alph) + sigma)

    if sw is False:

        taudm *= xsw2pw
        taudm_minus *= xsw2pw
        taudm_plus *= xsw2pw

        tau_outer_extend_minus *= xsw2pw
        tau_outer_extend_plus *= xsw2pw
     
    return taudm, taudm_minus, taudm_plus, tau_outer_extend_minus, tau_outer_extend_plus

  



# mwprop.ne2001p v1.0 Jan 2024

from __future__ import print_function
# revised 2020 Feb 11 to remove any routines that use Fortran NE2001
# revised 2019 Dec 31 - 2020 Jan 1 to be a proper python package
# original 02 Apr 2017
# Utilities

import numpy as np
from math import *
from scipy import stats
from matplotlib.pyplot import rc, figure, axes, axis, xscale, yscale, \
                       plot, fill_between, xlabel, ylabel, annotate, \
                       title, legend, grid, savefig, show, close 
import sys

# Import useful constants ... get them all (not that many)
# Note other routines in this subpackage see bare constants (e.g. KDM) 
# but if this subpackage is imported in ipython as
# from mwprop.scattering_functions import iss_mw_utils2020p as ut
# need ut.KDM to see the value of KDM.

from mwprop.get_constants_waveprop_scipy_version import * 

import mwprop.scattering_functions.scattering_functions2020 as sf 

input = input

#======================================================================`
def testit():
    print(KDM)
#======================================================================`

def theta_iso_at_RF(RF, SMiso, DEFFSM, scattype,  si=11./3.):
   """
   Calculates the isoplanatic angle in milliarcseconds at the specified RF.
   Assumes a Kolmogorov scaling (si=11/3).

   Input:  RF       GHz
           SM_iso   standard SM uits
           DEFFSM   kpc
           si = wavenumber spectral index for electron-density (e.g. 11/3)
   Output: 
           THETA_ISO            mas at 1 GHz
           THETA_ISO_RF         mas at RF

   Notes:
   For strong scattering:
   From function THETA_ISO in scattering98.f (part of the NE2001 code package):
   \theta_{iso} = \delta r_s / d
                = \left [
                     (\lambda r_e)^2 f_{\alpha} SM_{iso}
                  \right ]^{1/\alpha}
   where 
         \lambda = EM wavelength (cm)
         re = classical electron radius (cgs)
         \alpha = 5/3 for Kolmogorov case.
         SM_{iso} = \int_0^d ds s^{\alpha} \cnsq
            so SM_{iso} does not have the units of scattering
            measure, but rather units of SM x Length^{\alpha}
 
         f_{\alpha} = 
           8\pi^2 \Gamma(1-\alpha/2) / [\alpha 2^{\alpha} \Gamma(1+\alpha/2)]
           for \alpha = 5/3, f_{\alpha}= 88.3
 
   theta_log_radian =
       13.287                                ! 0.6*log10(30cm*r_e)
     + 1.2 * np.log10(nu)
     - 1.1676                                ! 0.6*log10(f_alpha)
     - 0.6 * np.log10(smiso)
     - 34.383                                ! 1.6 * alog10(kpc)
     + 8.                                    ! -(20/3)*log(100)
   theta_log_microarcsec =
       theta_log_radian + 11.314425        ! 11.314425=alog10(microarsec/rad)

   Weak scattering:   the scattered source size is the Fresnel scale / distance.
       theta_iso = lambda / Deff

   """
 
   if scattype == 'strong' or scattype == 'transition':
      xiso = si / (si-2.) - 1.

      theta_log_radian = \
        13.287\
        - 1.1676 \
        - 0.6 * np.log10(SMiso) \
        - 34.383 \
        + 8.                                     
      theta_log_microarcsec = \
          theta_log_radian + 11.314425        # 11.314425=alog10(microarsec/rad)
      THETA_ISO = 10.**theta_log_radian / mas
      THETA_ISO_RF = THETA_ISO * RF**xiso

   if scattype == 'weak':       # NEEDS CHECKING
      xiso = -0.5 
      rF_1GHz = np.sqrt(wavelen1GHz * DEFFSM * kpc / (2.*pi))
      THETA_ISO = (wavelen1GHz / rF_1GHz) / mas     # mas at 1 GHz 
      THETA_ISO_RF = THETA_ISO * RF**xiso

   return THETA_ISO, THETA_ISO_RF


#======================================================================`

def scale_diss_params_to_RF(rfratio, TAU, THETA_X, si=11./3.):

   """
   Scales relevant parameters outputed by NE2001 to the designated RF 
   assuming a Kolmogorov scaling (si=11/3).

   Input:  
           rfratio = (ratio of RF for output values) / (RF for input values)
           TAU          ms
           THETA_X      mas
           si = wavenumber spectral index for electron-density  
   Output: 
           TAU_RF           ms at RF  
           THETA_X_RF       mas at RF
   """

   xsbw = 2.*si / (si-2.)
   xtau = -xsbw 
   xthetax = -xsbw/2.

   TAU_RF = TAU * rfratio**xtau  
   THETA_X_RF = THETA_X * rfratio**xthetax

   return TAU_RF, THETA_X_RF

#======================================================================`

def scale_NE2001_output_to_RF(RF,TAU,SBW,SCINTIME,THETA_G,THETA_X,si=11./3.):
   """
   Scales relevant parameters outputed by NE2001 to the designated RF 
   assuming a Kolmogorov scaling (si=11/3).

   Input:  
           TAU          ms
           SBW          MHz
           SCINTIME     sec
           THETA_G      mas
           THETA_X      mas
           RF           GHz
           si = wavenumber spectral index for electron-density  
   Output: 
           TAU_RF           ms at RF  
           SBW_RF           MHz at RF  
           SCINTIME_RF      sec at RF  
           THETA_G_RF       mas at RF  
           THETA_X_RF       mas at RF
   """

   xsbw = 2.*si / (si-2.)
   xtau = -xsbw 
   xscintime = xsbw/2. - 1.
   xthetag = -xsbw/2.
   xthetax = xthetag

   TAU_RF = TAU * RF**xtau  
   SBW_RF = SBW * RF**xsbw   
   SCINTIME_RF = SCINTIME * RF**xscintime   
   THETA_G_RF = THETA_G * RF**xthetag  
   THETA_X_RF = THETA_X * RF**xthetax

   return TAU_RF, SBW_RF, SCINTIME_RF, THETA_G_RF, THETA_X_RF

#======================================================================`

def ggpdf(x, rms1, rms2):
   """
   Calculates gamma-gamma PDF for combined diffractive and refractive 
   scintillations given in Eq. 20 of Prokes, 
   'Modeling of Atmospheric Turbulence Effect on Terrestrial FSO Link'

   """
   from math import gamma
   from scipy.special import kv
   from scipy.special import kve
   from scipy.special import gammaln

   a = 1./rms1**2
   b = 1./rms2**2

   e1 = (a+b)/2. 
   nu = a-b
   arg1 = 2. * np.sqrt(a*b*x)   

   #coeff1 = (2. * (a*b)**e1) / (gamma(a) * gamma(b)) 
   coeffln = np.log(2.) + e1*np.log(a*b) - gammaln(a) - gammaln(b)
   kve = kve(nu, arg1)
   pdfln = coeffln + (e1-1.)*np.log(x) + np.log(kv(nu, arg1))
   #pdf = coeff1 * x**(e1-1) * kv(nu, arg1)
   pdf = np.exp(pdfln)
   return pdf

#======================================================================`

def mriss_goodman_narayan(si, phif):
   """
   Returns the RISS modulation index using equation 3.1.5 of
   Goodman & Narayan 1985.

   Input: 
      si = beta = wavenumber index of electron density power spectrum
      e.g. beta = 11/3 for Kolmogorov case

      Applicable for 2 < si < 4.

      phif = rms phase difference on the Fresnel scale
             with Fresnel scale for plane-wave incidence on a phase screen
             at distance d: rf = \lambda d / 2\pi. 

             The "u" parameter is calculated from phif as 
             u = phif^{2/(\beta -2} in the evaluation.
   Output: 

      mriss 

   """
   from math import gamma

   si2 = si-2.
   si4 = 4.-si
   si6 = 6.-si
   Crsq = 2.**(-si*si4/si2) * gamma(si/2.)*gamma(si6/si2) / gamma(si6/2.)

   u = phif**(2./si2)

   mriss = np.sqrt(Crsq) * 2.**(si4/si2**2) * u**(-si4)

   return mriss

#======================================================================`

def lnorm_from_mean_rms(x, mean, rms):
   """
   Calculates lognorm pdf for specificed x range and mean, rms of x.
   Note:

   mean = exp(mu+sigma**2/2) 
   median = exp(mu) = 1.05
   variance = (e^{sigma^2}-1) e^{2*mu+sigma^2}

   where mu, sigma are standard parameters for the log-normal pdf 
   """

   sigma = np.sqrt(np.log(1 + (rms/mean)**2))
   mu = np.log(mean) - sigma**2/2.

   # lognorm.pdf seems to give the wrong input rms
   # even though it has unit area and unit mean when
   # I input mean = 1. and rms = 0.3
   # it gives lognorm.stats(sigma, mu) = 1., 0.313
   #pdfx = lognorm.pdf(x, sigma, mu, 1.)

   # so calculate directly: this gives correct input mean and rms:

   pdfx = (x * sigma * np.sqrt(2.*pi))**(-1.) * np.exp(-(np.log(x)-mu)**2/(2.*sigma**2))

   return mu, sigma, pdfx


#======================================================================`

def calculate_fresnel_scale_and_mod_indices(RF, DEFFSM, THETA_X_RF, si=11./3.):
    """
    Calculates phiF and modulation indices.
    si = index of wavenumber spectrum for ne 
    """

    wavelength = c / (RF*GHz)
    rF = np.sqrt(DEFFSM *kpc * wavelength / (2.*pi))    # Fresnel scale cm

    phiF = np.sqrt(2.) * \
       ( (pi*THETA_X_RF*mas * rF / wavelength) / (2.*np.sqrt(np.log(2.))) )**((si-2.)/2.) 

    u = phiF**(2./(si-2.))

    """
    Modulation indices for different regimes of u:
    use definitions of md, mr that are continuous across the 
    weak/strong boundary. Do so by using an equipartition approach for 
    u << 1 but one that asymptotically gives
    reasonable modulation indices md = 1 and mr = (2u)^{-1/2}
    """

    mdp = np.sqrt(2.)*u / np.sqrt(1. + 2.*u**2)

    mrp = np.sqrt(2.) * u * (2*u)**(si-4.) / np.sqrt((2*u)**(2.*(si-4.)) + 2.*u**2)

    mgp = np.sqrt((1.+mdp**2)*(1.+mrp**2)-1.)
    if u <= 1.:     # no distinction between r,d
       lg = rF
       ld = rF
       lr = rF
       scattype = 'weak'
    if u > 1.:
       ld = rF/u
       lr = rF*u
       if u <= 2:
          scattype = 'transition'
       else:
          scattype = 'strong' 
    return rF, ld, lr, phiF,  u,  mgp, mdp, mrp, scattype 

#======================================================================`

def calculate_diss_bandwidth_factor(sbw, bw):
   """
   Calculates attentuation of DISS modulation index by bandwidth.
   Input:
      sbw = scintillation bandwidth
      bw = receiver bandwidth

      note sbw, bw must have same units
   Output:
      bwfactor
   """
   eta_nu = 0.3
   ndiss = 1.+eta_nu * bw / sbw 
   bwfactor = 1./np.sqrt(ndiss)
   return ndiss, bwfactor

#======================================================================`

def calculate_diss_source_size_factor(theta_source, theta_iso):
   """
   Calculates attentuation of DISS modulation index by source size.
   Input:
      theta_source = source size (angular units)
      theta_iso = isoplanatic size (angular units)

      note theta_source, theta_iso must  have same units
   Output:
      source_size_factor
   """
   #source_size_factor = min(1., theta_iso/theta_source)
   # changed 2016 July 5
   source_size_factor = (1. + (theta_source/ theta_iso)**2)**(-1./2.) 
   return source_size_factor

#======================================================================`

def invert_function(x, func, level):
   """
   Finds the value of x at which func = level.
   Interpolation is done.  
   """
   xlevel = np.interp(level, func, x)
   return xlevel


#======================================================================`

def calc_pdfg_for_xgal_los(RF, BW, RF_input, TAU, THETA_X, Deff, SM, 
    theta_source=1.e-6, dg=1.e-5, gmin=1.e-5, gmax=30., si=11./3.):

    """
    Calculates the PDF and CDF for the  ISS modulation for the 
    a line of sight characterized by TAU, THETA_X, Deff, SM
    and for the specified RF and bandwidth.

    RF              Radio frequency of evaluated PDF, CDF       GHz
    BW              Bandwidth                                   MHz
    RF_input        Radio frequency of input ISS values         GHz
    TAU             Pulse broadening time                        ms
    THETA_X         Angular diameter of extragalactic 
                    source caused by MW scattering              mas
    Deff            Effective distance to MW scattering region  kpc
    SM              Scattering measure                          kpc m^{-20/}
    theta_source    Source size as seen by ISM                  mas
    dg,gmin,gmax    Sample interval and Range for gain g in pdf,cdf
    si              Spectral index of electron density
                    wavenumber spectrum (11/3 for Kolmogorov)

    Note d_xgal_kpc has little effect; it's mainly to ensure that the NE2001
    code integrates out to its maximum distance.
    """

    C1 = 1.16                       # factor in SBW-TAU relation

    # Scale input parameters to RF of interest if necessary:
    if RF == RF_input:
        TAU_RF = TAU
        THETA_X_RF = THETA_X
    else:
        rfratio = RF / RF_input
        TAU_RF, THETA_X_RF = \
         scale_diss_params_to_RF(rfratio, TAU, THETA_X, si=si)

    rF, ld, lr, phiF, u, mgp, mdp, mrp, scattype = \
       calculate_fresnel_scale_and_mod_indices(RF, Deff, THETA_X_RF,si=si)

    THETA_ISO, THETA_ISO_RF = theta_iso_at_RF( \
                              RF, SM, Deff, scattype, si=si)

    SCINTIME_RF = sf.scintime(SM, RF, vperp=100)        # for 100 km/s

    SBW_RF = 1.e-3 * C1 / (2. * pi * TAU_RF)            # MHz 

    nu_transition = sf.transition_frequency_from_obs(RF, SBW_RF, si=si) 

    ndiss, bandwidth_factor = calculate_diss_bandwidth_factor(SBW_RF, BW)

    source_size_factor =  \
             calculate_diss_source_size_factor(theta_source, THETA_ISO_RF)

    # set modulation indices using source size and bandwidth factors:

    gmean = 1.
    mr = mrp
    md = mdp * source_size_factor * bandwidth_factor


    if scattype == 'strong':
       stypelab = 'S'
       mg = np.sqrt(md**2 + mr**2 + (md*mr)**2) # ok for strong scattering
       mdtrans = 0.5 + 0.25*mr

       if md <= mdtrans:
         pdftype = 'log-normal'
         pdftypelab = 'LN'

       if md > mdtrans and mr > 0.1:
         pdftype = 'gamma-gamma'
         pdftypelab = 'GG'

       if md > mdtrans and mr <= 0.1:
         pdftype = 'chi-square'
         pdftypelab = 'X'
         dof_g = 2. / mg**2         # DoF for chi^2 pdf 

    if scattype == 'transition':    # 
         stypelab = 'T'
         pdftype = 'gamma-gamma'
         pdftypelab = 'GG'

    if scattype == 'weak':      # log-normal PDF 
         stypelab = 'W'
         mg = mgp * source_size_factor
         sigma = np.sqrt(np.log(1.+mg**2))
         mu = -sigma**2/2.
         pdftype = 'log-normal'
         pdftypelab = 'LN'

    # pdf and cdf:
    gvec = np.arange(gmin, gmax+dg, dg)
    wvec = np.ones(np.size(gvec))
    wvec[0] = wvec[-1] = 0.5

    if pdftype == 'log-normal':
       pdflabel = r'$\rm LN$'
       muln, sigmaln, gpdf = lnorm_from_mean_rms(gvec, gmean, mg)
       gcdf = (np.cumsum(wvec*gpdf) + lnorm_from_mean_rms(dg/2., gmean, mg)[2]) * dg

    if pdftype == 'chi-square':
       pdflabel = r'$\chi^2_{\rm n}$'
       loc = 0.
       scale = 1. / dof_g       # gives pdf of reduced chi-square quantity
       gpdf = chi2.pdf(gvec, dof_g, loc, scale) 
       gcdf = (np.cumsum(wvec*gpdf) + chi2.pdf(dg/2., dof_g, loc, scale)) * dg
       
    if pdftype == 'gamma-gamma':
       pdflabel = r'$\gamma\gamma}$'
       gpdf = ggpdf(gvec, md, mr)
       gcdf = (np.cumsum(wvec*gpdf) + ggpdf(dg/2., md, mr)) * dg

    # find median and probability ranges for g:
    gmedian =  invert_function(gvec, gcdf, 0.5)

    # Define new dictionary that contains output values:

    Disskeys = \
        np.array(['RF', 'tau', 'SM', 'tau_rf', 'theta_x', 'theta_x_rf', 'rFresnel',
       'ld', 'lr', 'phiF', 'u', 'scattype', 'mr', 'md', 'mdp', 'mg',
       'gmedian', 'stypelab', 'pdftype',  'pdflabel', 'gvec', 'gpdf',
       'gcdf', 'nu_transition', 'theta_iso', 'theta_iso_rf', 'sbw_rf',
       'scintime_rf', 'bandwidth_factor', 'source_size_factor'])

    Dissvals = \
        list([RF, TAU, SM, TAU_RF, THETA_X, THETA_X_RF, rF, ld, lr, phiF, u, scattype, mr, md, mdp, mg, gmedian, stypelab, pdftype, pdflabel, gvec, gpdf, gcdf, nu_transition, THETA_ISO, THETA_ISO_RF, SBW_RF, SCINTIME_RF, bandwidth_factor, source_size_factor]) 

    Dissunits = np.array([
        'GHz', 'ms', 'kpc m^{-20/3}', 'ms', 'mas', 'mas', 'cm', 'cm', 'cm', 'rad', '', '', '', '', '', '', '', '', '', '', '', '', '', '', 'GHz', 'mas', 'mas', 'MHz', 's', '', '' ])

    Dissdesc = np.array([
        'RF of output values', 'TAU input', 'SM input', 'TAU @RF', 'XGalFWHM in',  'XgalFWHM @RF', 'Fresnel scale', 'Diffraction scale', 'Refraction scale', 'Fresnel phase', 'U parameter', 'Scattering regime', 'Refraction mod index', 'Diffraction mod index', 'Diffraction mod index (pt source)', 'Mod index (total)', 'Median mod', 'Scattering label', 'PDF type', 'PDF label', 'Gain vector', 'Gain PDF', 'GAIN CDF', 'Transition RF', 'Isoplanatic angle input', 'Isoplanatic angle @RF', 'DISS bandwidth @RF', 'DISS time scale @RF', 'Bandwidth factor', 'Source size factor'])

    Dissv, Dissu, Dissd = {}, {}, {}
    for n, key in enumerate(Disskeys):
       Dissv[key] = Dissvals[n]
       Dissu[key] = Dissunits[n]
       Dissd[key] = Dissdesc[n]

    """
    Can delete these lines once Dissv,u,d are confirmed accurate
    D_iss = {}
    D_iss['RF'] = RF
    D_iss['tau'] = TAU
    D_iss['SM'] = SM
    D_iss['tau_rf'] = TAU_RF
    D_iss['theta_x'] = THETA_X
    D_iss['theta_x_rf'] = THETA_X_RF
    D_iss['rFresnel'] = rF
    D_iss['ld'] = ld
    D_iss['lr'] = lr
    D_iss['phiF'] = phiF
    D_iss['u'] = u 
    D_iss['scattype'] = scattype
    D_iss['mr'] = mr
    D_iss['md'] = md
    D_iss['mdp'] = mdp
    D_iss['mg'] = mg
    D_iss['gmedian'] = gmedian
    D_iss['stypelab'] = stypelab
    D_iss['pdftypelab'] = pdftypelab
    D_iss['pdflabel'] = pdflabel
    D_iss['gvec'] = gvec
    D_iss['gpdf'] = gpdf
    D_iss['gcdf'] = gcdf
    D_iss['nu_transition'] = nu_transition
    D_iss['theta_iso'] = THETA_ISO 
    D_iss['theta_iso_rf'] = THETA_ISO_RF 
    D_iss['sbw_rf'] = SBW_RF
    D_iss['scintime_rf'] = SCINTIME_RF
    D_iss['bandwidth_factor'] = bandwidth_factor
    D_iss['source_size_factor'] = source_size_factor 
    return D_iss, Disskeys, Dissvals 
    """

    return Dissv, Dissu, Dissd

#======================================================================`

def main(l, b, RF, BW, dxgal_mpc, theta_source, SM, DMmodel, doplot=False):
    # setup is so that a plot is done if run as a program

    # output line is appended to file xgal_mw_iss_output.txt
    # some values and arrays (gvec, gpdf, gcdf) are saved into an npz file,
    # xgal_mw_iss_save.npz


    print("input = ", l,b,RF,BW,dxgal_mpc,theta_source)

    """ 
    Evaluates DISS and RISS modulation indices and probabilities for the net gain
    given a direction (l,b), frequency \nu, bandwidth B, and extragalactic source size in milliarcseconds.
    """
    outfile = 'iss_mw_pdf_package.txt'
    outfile2 = 'iss_mw_pdf_package_lb_granges.txt'
    fout = open(outfile, 'a')
    gout = open(outfile2, 'a')
    hout = open('junkfile_iss_package', 'a')

    D_iss = calc_pdfg_for_xgal_los(
                  RF, BW, RF_input, TAU, THETA_X, Deff, SM,
                  theta_source=1.e-6, dg=1.e-5, gmin=1.e-5, gmax=30.)

    gvec = D_iss['gvec']
    gpdf = D_iss['gpdf']
    gcdf = D_iss['gcdf']
    NU_T = D_iss['nu_transition']
    SM = D_iss['SM']
    THETA_X_RF = D_iss['theta_x_rf']
    THETA_ISO = D_iss['theta_iso']
    THETA_ISO_RF = D_iss['theta_iso_rf']
    SBW_RF = D_iss['sbw_rf'] 
    phiF = D_iss['phiF'] 
    rF = D_iss['rFresnel']
    lr = D_iss['lr']
    ld = D_iss['ld']
    bandwidth_factor = D_iss['bandwidth_factor'] 
    source_size_factor = D_iss['source_size_factor'] 
    mr = D_iss['mr']
    md = D_iss['md']
    mdp = D_iss['mdp']
    mg = D_iss['mg']
    gmedian = D_iss['gmedian']
    stypelab = D_iss['stypelab'] 
    pdftypelab = D_iss['pdftypelab'] 
    pdflabel = D_iss['pdflabel'] 
    scattype = D_iss['scattype']
    SCINTIME_RF =  D_iss['scintime_rf'] 

    p80 = 0.8
    g80low = invert_function(gvec, gcdf, (1.-p80)/2.)
    g80high = invert_function(gvec, gcdf, (1.+p80)/2.)
      
    p90 = 0.9
    g90low = invert_function(gvec, gcdf, (1.-p90)/2.)
    g90high = invert_function(gvec, gcdf, (1.+p90)/2.)
      
    p98 = 0.98
    g98low = invert_function(gvec, gcdf, (1.-p98)/2.)
    g98high = invert_function(gvec, gcdf, (1.+p98)/2.)
      
    p998 = 0.998
    g998low = invert_function(gvec, gcdf, (1.-p998)/2.)
    g998high = invert_function(gvec, gcdf, (1.+p998)/2.)
      
    p9998 = 0.9998

    g9998low = invert_function(gvec, gcdf, (1.-p9998)/2.)
    g9998high = invert_function(gvec, gcdf, (1.+p9998)/2.)
      
    print(\
       "%6.1f %5.1f %4.1f %4.1f %7.1f %7.1f %6.3e %6.2f %4.1e %5.2f \
       %8.4e %8.2f %5.2f %5.2f %5.2f %4.3e %6.3e %6.3e %6.3e %5.3f \
       %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f \
       %2s %2s"\
       %(l, b, RF, BW, NU_T, DMmodel, THETA_ISO_RF*1000., THETA_X_RF, \
         SM, Deff, SBW_RF, phiF, np.log10(rF), np.log10(lr), np.log10(ld), \
         bandwidth_factor, mr, md, mg, gmedian,\
         g80low, g80high,g90low, g90high, g98low, g98high, \
         g998low, g998high,g9998low, g9998high,\
         stypelab, pdftypelab), file=fout) 

    print( "DM = ", DMmodel)
    print( "SM = ", SM)
    print( "Transition frequency = ", NU_T)
    print( "Theta xgal = ", THETA_X_RF, " mas")
    print( "Effective distance = ", Deff)

    print( "phiF = ", phiF, " md = ", md, " mr = ", mr, " mg = ", mg, " SBW = ", SBW_RF, " bw_fac = ", bandwidth_factor)

    print("%6.1f %5.1f %4.1f %4.1f %7.1f %7.1f  %4.1e %5.2f %7.1f %4.1f %4.1f %4.1f %4.3e  %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f %5.3f"%(l,b,RF,BW, NU_T, DMmodel, SM, Deff, phiF, np.log10(rF), np.log10(lr), np.log10(ld), bandwidth_factor,  mg, gmedian, g80low, g80high, g90low, g90high, g98low, g98high,g998low, g998high,g9998low, g9998high), file=gout) 
       
    fout.close()
    gout.close()
    hout.close()

    # save key output
    savefile = 'iss_mw_pdf_package_save'
    np.savez(savefile, l=l, b=b, RF=RF, BW=BW, dxgal_mpc=dxgal_mpc, theta_source=theta_source, gvec=gvec, gpdf=gpdf, gcdf=gcdf, scattering=scattype, pdftype=pdflabel, mdp=mdp, md=md, mr=mr, mg=mg, gmedian=gmedian, g80low=g80low, g80high=g80high, g90low=g90low, g90high=g90high, g98low=g98low, g98high=g98high,g998low=g998low, g998high=g998high, g9998low=g9998low, g9998high=g9998high, DMmodel=DMmodel, Deff=Deff, SM=SM, SBW_RF = SBW_RF, phiF=phiF, bandwidth_factor=bandwidth_factor)

    if doplot:
      # Plotting:

      # not sure if these rc specs are working
      rc('font', family='serif')
      rc('text', usetex=True)
      rc('font', serif='cm')

      # PDF
      fig = figure()
      ax = fig.add_subplot(111)
      xlabel(r'$\rm g = g_{\rm d} \  g_{\rm r}$', fontsize=15)
      ylabel(r'$\rm Probability \ Density \ \ f_g(g)$', fontsize=15)
      plot(gvec, gpdf, 'k-', label = scattype + ',   ' + pdflabel)
      yscale('log')
      axis(xmin = -0.05)
      axis(ymin = 1.e-10, ymax = 2.*gpdf.max())

      aa,bb,cc,dd = axis()
      plot((gmedian, gmedian), (cc,dd), 'k', dashes=[20,5], lw=3)
      plot((g90low, g90low), (cc,dd), 'k', dashes=[20,3], lw=1)
      plot((g90high, g90high), (cc,dd), 'k', dashes=[10,3], lw=1)

      annotate(r'$\rm l,b \ = \  %4.1f^{\circ}, %4.1f^{\circ} \ \ \ DM_{NE2001} = %6.1f \ pc\ cm^{-3}  $'%(l,b,DMmodel), xy=(0.5, 1.03), xycoords='axes fraction', ha='center', fontsize=16)

      xanno = 0.3
      xanno = 0.95
      ha = 'right'
      yanno = 0.925
      dyanno = 0.06
      fontsize = 13
      annotate(r'$\rm \nu  =   %4.2f \ GHz \ \ \ \  B  =  %4.2f\ GHz \ \ \ \ \phi_{\rm F}   =  %6.1f \ rad $'%(RF, BW/1000., phiF), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      yanno -= dyanno

      annotate(r'$\rm  r_{\rm F} =   10^{%4.1f} \ cm\ \ \ \  l_d  =  10^{%4.1f}\ cm \ \ \ \ l_r   =  10^{%4.1f} \ cm$'%(np.log10(rF), np.log10(ld), np.log10(lr)), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      yanno -= dyanno

      annotate(r'$\rm  \Delta\nu_d  =   {%4.1f} \ MHz\ \ \ \  \Delta t_d=  {%4.1f}\ s\ \ \ \ D_{\rm eff}   =  %4.2f \ kpc$'%(SBW_RF, SCINTIME_RF, Deff), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      yanno -= dyanno

      annotate(r'$\rm  \theta_{\rm iso}  =   {%4.1f} \ \mu as\ \ \ \  \theta_{\rm MW} =  {%4.1f}\ mas\ $'%(THETA_ISO*1000., THETA_X_RF), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      legend(loc=(0.7, 0.6))
      show()
      savefig('xgal_mw_iss_pdf_l_' + str(l) + '_b_' + str(b) + '_rf_' + str(RF) + '_bw_' + str(BW)+ '.pdf')

      # CDF
      fig = figure()
      ax = fig.add_subplot(111)
      rc('xtick', labelsize=16)
      rc('ytick', labelsize=16)

      xlabel(r'$\rm g = g_{\rm d} \  g_{\rm r}$', fontsize=15)
      ylabel(r'$\rm CDF\ \ F_g(g)$', fontsize=15)
      plot(gvec, gcdf, 'k-', label = scattype + ',   ' + pdflabel)
      yscale('log')
      axis(xmin = -0.05)
      axis(ymin = 1.e-3, ymax=2.)

      aa,bb,cc,dd = axis()
      plot((gmedian, gmedian), (cc,dd), 'k', dashes=[20,5], lw=3)
      plot((g90low, g90low), (cc,dd), 'k', dashes=[20,3], lw=1)
      plot((g90high, g90high), (cc,dd), 'k', dashes=[10,3], lw=1)

      annotate(r'$\rm l,b \ = \  %4.1f^{\circ}, %4.1f^{\circ} \ \ \ DM_{NE2001} = %6.1f \ pc\ cm^{-3}  $'%(l,b,DMmodel), xy=(0.5, 1.03), xycoords='axes fraction', ha='center', fontsize=16)

      xanno = 0.3
      xanno = 0.95
      ha = 'right'
      yanno = 0.225
      dyanno = 0.06
      fontsize = 13
      annotate(r'$\rm \nu  =   %4.2f \ GHz \ \ \ \  B  =  %4.2f\ GHz \ \ \ \ \phi_{\rm F}   =  %6.1f \ rad $'%(RF, BW/1000., phiF), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      yanno -= dyanno

      annotate(r'$\rm  r_{\rm F} =   10^{%4.1f} \ cm\ \ \ \  l_d  =  10^{%4.1f}\ cm \ \ \ \ l_r   =  10^{%4.1f} \ cm$'%(np.log10(rF), np.log10(ld), np.log10(lr)), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      yanno -= dyanno

      annotate(r'$\rm  \Delta\nu_d  =   {%4.1f} \ MHz\ \ \ \  \Delta t_d=  {%4.1f}\ s\ \ \ \ D_{\rm eff}   =  %4.2f \ kpc$'%(SBW_RF, SCINTIME_RF, Deff), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      yanno -= dyanno

      annotate(r'$\rm  \theta_{\rm iso}  =   {%4.1f} \ \mu as\ \ \ \  \theta_{\rm MW} =  {%4.1f}\ mas\ $'%(THETA_ISO*1000., THETA_X_RF), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      legend(loc=(0.7, 0.8))
      show()
      savefig('xgal_mw_iss_cdf_l_' + str(l) + '_b_' + str(b) + '_rf_' + str(RF) + '.pdf')

      print( "Plotting Complementary CDF")
      doextremes_default = True 
      #doextremes = bool(input('Plot extreme value examples? [%s]: '%(doextremes_default)) or str(doextremes_default))
      

      # CCDF = 1 - CDF
      fig = figure()
      ax = fig.add_subplot(111)
      rc('xtick', labelsize=16)
      rc('ytick', labelsize=16)
      xlabel(r'$\rm g = g_{\rm d} \  g_{\rm r}$', fontsize=16)
      ylabel(r'$\rm Complementary \ CDF\ \ 1-F_g(g)$', fontsize=16)
      plot(gvec, 1.-gcdf, 'k-', label = scattype + ',   ' + pdflabel)
      yscale('log')
      #yticklocs, yticklabs = calcticks.calcticks_log(tlog1=-10, tlog2=0, dt = 1)
      #ax.set_yticks(yticklocs)
      #ax.set_yticklabels(yticklabs)
      axis(xmin = -0.05)
      axis(ymin = 1.e-10, ymax=2.)

      aa,bb,cc,dd = axis()

      # median dashed line:
      plot((gmedian, gmedian), (cc,0.5), 'k', dashes=[20,5], lw=1)

      idx = np.where((g90low <= gvec) & (gvec <=g90high))
      # idx contains a lot of points (for sample interval used on gvec) 
      # so the plot file ends up being very large.   Therefore just use 
      # the edges and midpoint. 

      i00 = idx[0][0]
      i0half = idx[0][np.int(np.size(idx)/2)]
      i01 = idx[0][-1]
      gfill = ((gvec[i00], gvec[i0half], gvec[i01]))
      gccdffill = ((1.-gcdf[i00], 1.-gcdf[i0half], 1.-gcdf[i01]))

      if fill90:
         fill_between(gfill, gccdffill, edgecolor='white', facecolor='cyan', alpha=0.2)
      else:
         plot((g90low, g90low), (cc, 0.95), 'k', dashes=[20,3], lw=1)
         plot((g90high, g90high), (cc,0.05), 'k', dashes=[10,3], lw=1)

      xanno = 0.3
      xanno = 0.95
      ha = 'right'
      yanno = 0.925
      dyanno = 0.06
      fontsize = 13
      annotate(r'$\rm l,b \ = \  %4.1f^{\circ}, %4.1f^{\circ} \ \ \ DM_{NE2001} = %6.1f \ pc\ cm^{-3}  $'%(l,b,DMmodel), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      yanno -= dyanno
      annotate(r'$\rm \nu  =   %4.2f \ GHz \ \ \ \  B  =  %4.2f\ GHz \ \ \ \ \phi_{\rm F}   =  %6.1f \ rad $'%(RF, BW/1000., phiF), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      yanno -= dyanno

      annotate(r'$\rm  r_{\rm F} =   10^{%4.1f} \ cm\ \ \ \  l_d  =  10^{%4.1f}\ cm \ \ \ \ l_r   =  10^{%4.1f} \ cm$'%(np.log10(rF), np.log10(ld), np.log10(lr)), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      yanno -= dyanno

      if SBW_RF > 1.:
           annotate(r'$\rm  \Delta\nu_d  =   {%4.1f} \ MHz\ \ \ \  \Delta t_d=  {%5.0f}\ s\ \ \ \ D_{\rm eff}   =  %4.2f \ kpc$'%(SBW_RF, SCINTIME_RF, Deff), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)
      else:
           annotate(r'$\rm  \Delta\nu_d  =   {%4.2f} \ MHz\ \ \ \  \Delta t_d=  {%5.0f}\ s\ \ \ \ D_{\rm eff}   =  %4.2f \ kpc$'%(SBW_RF, SCINTIME_RF, Deff), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)
      yanno -= dyanno

      annotate(r'$\rm  \theta_{\rm iso}  =   {%4.1f} \ \mu as\ \ \ \  \theta_{\rm MW} =  {%4.1f}\ mas\ $'%(THETA_ISO*1000., THETA_X_RF), xy=(xanno, yanno), xycoords='axes fraction', ha=ha, fontsize=fontsize)

      if doextremes:  # plot points for specified probability levels
           p3 = 1.e-3
           p8 = 1.e-8
           # note second arg needs to be an increasing function
           # so that is why it operates on gcdf instead of 1.-gcdf
           g3 = invert_function(gvec, gcdf, 1.-p3)
           g8 = invert_function(gvec, gcdf, 1.-p8)

           plot(g3, p3, 'ro')
           plot(g8, p8, 'ro')

           aa,bb,cc,dd = axis()
           plot((g3, g3), (cc, p3), 'r:', lw=2)
           plot((g8, g8), (cc, p8), 'r:', lw=2)


           legend(loc=(0.675, 0.5))
      show()
      if fill90: 
         savefig('xgal_mw_iss_ccdf_l_' + str(l) + '_b_' + str(b) + '_rf_' + str(RF) + '_bw_' + str(BW) + '_fill'  + '.pdf')
      else:
         savefig('xgal_mw_iss_ccdf_l_' + str(l) + '_b_' + str(b) + '_rf_' + str(RF) + '_bw_' + str(BW) + '.pdf')
      input('hit return')
      close()

#======================================================================`

# Main

if __name__ == "__main__":

    fill90 = False
    fill90 = True

    narg=0
    print(sys.argv)
    for arg in sys.argv:
        #print narg, arg
        if  narg == 1: l = float(arg)
        if narg == 2: b = float(arg)
        if narg == 3: RF = float(arg) 
        if narg == 4: BW = float(arg) 
        if narg == 5: dxgal_mpc = float(arg) 
        if narg == 6: theta_source = float(arg) 
        if narg == 7: doplot = arg
        narg += 1
    narg -= 1
    print("narg = ", narg)

    if narg == 0:
        doplot = True
        l = float(input('enter Galactic longitude in degrees: '))
        b = float(input('enter Galactic latitude in degrees: '))
        RF = float(input('enter RF in GHz: '))
        BW = float(input('enter BW in MHz: '))
        dxgal_mpc=float(input('enter distance of extragalactic source (Mpc): '))
        theta_source=float(input('extragalactic source size (mas): '))

    RF_input = 1.
    TAU = 1.
    THETA_X = 1.
    Deff = 2.5
    SM = 10.**(-3.5)
    DMmodel = 100

    D_iss = calc_pdfg_for_xgal_los(
                  RF, BW, RF_input, TAU, THETA_X, Deff, SM,
                  theta_source=1.e-6, dg=1.e-5, gmin=1.e-5, gmax=30.)
    main(l, b, RF, BW, dxgal_mpc, theta_source, SM, DMmodel)




# mwprop v2.0 Jan 2026

"""
density.py

Returns the local electron density, fluctuation parameters, weights
and clump/void flags for the position (x, y, z)

Relacement for Fortran subroutine density_2001 


Comments from Fortran code density.NE2001.f
c final version of NE2001
c returns densities, F parameters and weights of the various components
c mods:
c 28 July 02:
c   put in 'if' statements in subroutine density_2001 so that
c   function calls are not done if particular weights
c   (wg1, wg2, etc.) are set to zero in gal01.inp
c   This is (a) cleaner and (b) much more efficient if
c   the clump or void component is flagged out.

Python version JMC 2020 Jan 28

Changes:
    2022 Jan 01: added functions density_2001_smallscale_comps and density_2001_smooth_comps
                 so that different LoS grids can be used to integrate different contributions
                 (to speed up calculations)
"""

from mwprop.nemod.config_nemod import *

from mwprop.nemod.density_components import * 
from mwprop.nemod.ne_lism import *
from mwprop.nemod.neclumpN_fast import *
from mwprop.nemod.nevoidN import *
from mwprop.nemod.ne_arms import *

script_path = os.path.dirname(os.path.realpath(__file__))

# ---------------------------------------------------------------------------

def density_2001(x, y, z, inds_relevant=None, verbose=False):  
    """
    (Edited) Comments from Fortran code:

     Returns seven components of the free electron density of the
     interstellar medium at Galactic location (x,y,z).

     input:
      x, y, z = galactocentric location (kpc)
          Right-handed coordinate system
          x is in l=90 direction
          y is in l=180 direction
          The sun is at (x,y,z) = (0,rsun,0)
     output:
      electron densities in cm^{-3}:
      ne1:    outer, thick disk
      ne2:    inner, thin disk (annular in form)
      nea:    spiral arms
      negc:   galactic center component
          nelism: local ISM component
          necN:   contribution from discrete 'clumps'
          nevN:   contribution from voids
       fluctuation parameters (one for each ne component):
          F1, F2, Fa, Fgc, Flism, FcN, FvN
       flags:
          whicharm: which of the 5 arms x,y,z is in (0 for interarm region)
             wlism: 1 if x,y,z is in any of the four LISM components
              wLDR: 1 if in LDR, 0 if not
              wLHB: 1 if in LHB, 0 if not
              wLSB: 1 if in LSB, 0 if not
            wLOOPI: 1 if in LoopI, 0 if not
          (nb: nelism is calculated according to LHB:LOOPI:LSB:LDR)
          hitclump: clump number that x,y,z is in (0 if none)
           hitvoid: void number that x,y,z is in (0 if none)
    25 May 2002
    based on routines from TC93 and test routines from 1999-2002 by JMC.
    """

    # Assumes model parameters have been put into global dictionaries
    # Initiate values in case components are flagged out

    ne1 = ne2 = nea = negc = nelism = necN = nevN = 0
    wlism=wldr=wlhb=wlsb=wloopI = hitclump = hitvoid = wvoid = whicharm = 0


    if wg1 == 1: 
        ne1, F1 = ne_outer(x,y,z)
    if wg2 == 1: 
        ne2, F2 = ne_inner(x,y,z)
    if wga == 1: 
        nea, Fa, whicharm = ne_arms_ne2001p(x,y,z, Ncoarse=Ncoarse)
    else:
        nea = Fa = 0.
    if wggc == 1:  
        negc, Fgc = ne_gc(x,y,z) 
    if wglism == 1: 
        nelism, Flism, wlism, wldr, wlhb, wlsb, wloopI = ne_lism(x,y,z)
    if wgcN == 1: 
        necN, FcN, hitclump = neclumpN(x,y,z, inds_relevant=inds_relevant)
    if wgvN == 1: 
        nevN, FvN, hitvoid, wvoid = nevoidN(x,y,z)

    if verbose:
        print('density: ', ne1, ne2, F1, F2)

    
    return ne1,ne2,nea,negc,nelism,necN,nevN, \
           F1, F2, Fa, Fgc, Flism, FcN, FvN, \
           whicharm, wlism, wldr, wlhb, wlsb, wloopI, \
           hitclump, hitvoid, wvoid


# ---------------------------------------------------------------------------

def density_2001_smallscale_comps(x, y, z, inds_relevant=None, verbose=False):  
    """
    (Edited) Comments from Fortran code:

     Returns four components of the free electron density of the
     interstellar medium at Galactic location (x,y,z).

     input:
      x, y, z = galactocentric location (kpc)
          Right-handed coordinate system
          x is in l=90 direction
          y is in l=180 direction
          The sun is at (x,y,z) = (0,rsun,0)
     output:
      electron densities in cm^{-3}:
      negc:   galactic center component
      nelism: local ISM component
      necN:   contribution from discrete 'clumps'
      nevN:   contribution from voids
      fluctuation parameters (one for each ne component):
          Fgc, Flism, FcN, FvN
      flags:
             wlism: 1 if x,y,z is in any of the four LISM components
              wLDR: 1 if in LDR, 0 if not
              wLHB: 1 if in LHB, 0 if not
              wLSB: 1 if in LSB, 0 if not
            wLOOPI: 1 if in LoopI, 0 if not
          (nb: nelism is calculated according to LHB:LOOPI:LSB:LDR)
          hitclump: clump number that x,y,z is in (0 if none)
           hitvoid: void number that x,y,z is in (0 if none)
    01 Jan 2022  based on fortran core from 25 May 2002
    """

    # Assumes model parameters have been put into global dictionaries
    # Initiate values in case components are flagged out

    negc = nelism = necN = nevN = 0
    wlism=wldr=wlhb=wlsb=wloopI = hitclump = hitvoid = wvoid = 0

    if wggc == 1:  
        negc, Fgc = ne_gc(x,y,z) 
    if wglism == 1: 
        nelism, Flism, wlism, wldr, wlhb, wlsb, wloopI = ne_lism(x,y,z)
    if wgcN == 1: 
        necN, FcN, hitclump, arg = neclumpN(x,y,z, inds_relevant=inds_relevant) # SKO added arg to output for debugging
    if wgvN == 1: 
        nevN, FvN, hitvoid, wvoid = nevoidN(x,y,z)

    #print('wgvN',wgvN,'hitvoid',hitvoid)
    #if verbose:
    #print('x,y,z,',x,y,z)
    #print('density: ', negc, nelism, necN, nevN) 

    if wgcN == 0:
        necN = 0
        FcN = 0
        hitclump = 0
    
    return negc,nelism,necN,nevN, \
           Fgc, Flism, FcN, FvN, \
           wlism, wldr, wlhb, wlsb, wloopI, \
           hitclump, hitvoid, wvoid

# ---------------------------------------------------------------------------

def density_2001_smooth_comps(x, y, z, verbose=False):  
    """
    Returns only the electron density for the smooth components.
    Utility: can be coarsely sampled to speed up computations.

    Output density is the sum ne1 + ne2 + nea at (x,y,z)

     input:
      x, y, z = galactocentric location (kpc)
          Right-handed coordinate system
          x is in l=90 direction
          y is in l=180 direction
          The sun is at (x,y,z) = (0,rsun,0)
     output:
      electron densities in cm^{-3}:
      ne1:    outer, thick disk
      ne2:    inner, thin disk (annular in form)
      nea:    spiral arms
    01 Jan 2022
    based on routines from NE2001 
    """

    # Assumes model parameters have been put into global dictionaries
    # Initiate values in case components are flagged out

    ne1 = ne2 = nea = 0
    F1 = F2 = Fa = 0
    whicharm = 0

    if wg1 == 1: 
        ne1, F1 = ne_outer(x,y,z)
    if wg2 == 1: 
        ne2, F2 = ne_inner(x,y,z)
    if wga == 1: 
        nea, Fa, whicharm = ne_arms_ne2001p(x,y,z, Ncoarse=Ncoarse)

    if verbose:
        print('density: ', ne1, ne2, F1, F2)

    # Define smooth components
    # Fsmooth defined so that 
    # Fsmooth*ne_smooth**2 = F1*ne1**2 + F2*ne2**2 + Fa*nea**2`
    wne1 = wg1*ne1
    wne2 = wg2*ne2
    wnea = wga*nea

    ne_smooth = wne1 + wne2 + wnea
    
    if ne_smooth > 0.:
        Fsmooth = (F1*(wne1)**2 + F2*wne2**2 + Fa*wnea**2) / ne_smooth**2
    else:
        Fsmooth = 0.

    return ne1,ne2,nea, F1, F2, Fa, whicharm, ne_smooth, Fsmooth

    """
    if verbose:
        return ne1,ne2,nea, \
               F1, F2, Fa, \
               whicharm, ne_smooth
    else:
        return ne_smooth
    """



#!/usr/bin/env python

# mwprop v2.0 Jan 2026

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

eval_NE2025 = False
eval_NE2001 = True

from mwprop import nemod
from mwprop.nemod.NE2001 import *


if __name__ == '__main__':

    # If 'explain' option is set, print out README file and exit.
    if '-e' in sys.argv or '--explain' in sys.argv:
        try:
            inp_file = (pkg_resources.files(nemod) / 'README.txt')
            with inp_file.open('rt') as f:
                template = f.read()
                print(template)
        except AttributeError:
            template = pkg_resources.read_text(nemod, 'README.txt')
            print(template)
        sys.exit()

    try:
        # parse command line arguments and execute
        parser = argparse.ArgumentParser(description='NE2001p (mwprop v2.0) Jan 2026')
        parser.add_argument('l',help='Galactic longitude (deg)')
        parser.add_argument('b',help='Galactic latitude (deg)')
        parser.add_argument('DM_D',help='DM (pc cm^{-3}) or distance (kpc)')
        parser.add_argument('ndir',help='ndir = 1 (DM->D), ndir = -1 (D->DM)')

        parser.add_argument('-a', '--analysis', action='store_true', 
            help='Do line of sight analysis')

        parser.add_argument('-p', '--plotting', action='store_true', 
            help='Do diagnostic plotting')

        parser.add_argument('-s', '--scattering', action='store_true', 
            help='Calculate scattering and scintillation parameters')

        parser.add_argument('-m', '--modern', action='store_true', 
            help='Modern output (turns off classic output like Fortran version)')

        parser.add_argument('-v', '--verbose', action='store_true', 
            help='Verbose output (classic output of Fortran version)')

        parser.add_argument('-b', '--debug', action='store_true', 
            help='debug output')
            
        args = parser.parse_args()


    except SystemExit:
        print('Try again with inputs')
        print('Use NE2001p.py -e to get explanation of code')
        sys.exit()

    ldeg = float(args.l)
    bdeg = float(args.b)
    dmd = float(args.DM_D)
    ndir = int(args.ndir)
    do_analysis = args.analysis
    do_plotting = args.plotting
    calc_scattering = args.scattering
    dmd_only = not calc_scattering
    verbose = args.verbose
    debug = args.debug
    modern = args.modern
    
    if modern:
        classic = False
    else:
        classic = True

    if do_plotting and calc_scattering:
        do_analysis = True # required for scattering plots

    Dn, Dv, Du, Dd = ne2001(ldeg, bdeg, dmd, ndir, 
        classic=classic, verbose=verbose, dmd_only=dmd_only, 
        do_analysis=do_analysis, plotting=do_plotting, debug=debug)

    if calc_scattering and do_plotting:
        output_dir = os.getcwd()+'/output_ne2001p/'
        if ndir >= 0:
            plot_dm_ne_cn2_vs_d(Dv, f25file = output_dir+'f25_dm2d_ne_dsm_vs_s.txt')
        if ndir < 0: # SKO added 12-15-23 to make output plotting work for both DM->D and D->DM
            plot_dm_ne_cn2_vs_d(Dv, f25file = output_dir+'f25_d2dm_ne_dsm_vs_s.txt')

        input('hit return')
        close('all')



#!/usr/bin/env python

# mwprop v2.0 Jan 2026


"""
los_diagnostics

Based on earlier los_plot.py and other code
Can vary Ncoarse = number of coarse samples along the LoS to test precision:
    Ncoarse  = 20           # as in Fortran code, produces discontinuities
    Ncoarse = 100           # reduces discontinuities, but no better than Ncoarse=50
    Ncoarse = 50            # reduces discontinuities

JMC 2024 Jan 3 
"""

from matplotlib.pyplot import suptitle
from numpy import *
import argparse
import sys,os

script_path = os.path.dirname(os.path.realpath(__file__))
basename = sys.argv[0].split('/')[-1].split('.')[0]         # for plot file name

def plot_ne_arms():
    # Plotting
    # ------------------------------------------------------------------
    # Three panel plot with n_e vs s, spiral arms, and arm number vs LoS
    # ------------------------------------------------------------------
    fig = figure()
    subplots_adjust(hspace=0.35, left=0.15, bottom=0.15, top = 0.85)
    suptitle(r'$\rm Output \ from \ los\_diagnostics.py$', fontsize=12)

    # Frame 1: ne from spiral arms vs distance
    ax1 = fig.add_subplot(211)
    plot(s25, ne25, 'k-', lw=6, label=r'$\rm n_e\, (total) $', alpha=0.7)
    plot(s, ne_lism_vec, 'r-', lw=3, label=r'$\rm {n_e}_{,lism} $')
    plot(s, nea_vec, '-', ms=2, label=r'$\rm n_e\, (arm) $')
    plot(s, ne1_vec, '-', ms=2, label=r'$\rm {n_e}_1 $')
    plot(s, ne2_vec, '-', ms=2, label=r'$\rm {n_e}_2 $')
    #plot(s, negc_vec, '-', ms=2, label=r'$\rm {n_e}_{gc} $')
    xlabel(r'$\rm d \ \ (kpc) $', fontsize=12)
    ylabel(r'$\rm n_e \ \ (cm^{-3}) $', fontsize=12)
    yy = axis()[3]
    axis(ymax = 1.35*yy)
    legend(loc=(0.025, 0.8), ncol=5, fontsize=9)
    if ndir > 0:
        title(r'$\rm DM = %5.1f\ pc\,cm^{-3} \ \longrightarrow \   D = %5.1f\ kpc \ \ \ \ \ \ \ \  l,\ b \ = \ %6.2f, \   %6.2f$'%(DM, d, ldeg, bdeg), fontsize=11)
    else:
        title(r'$\rm D = %5.1f\, kpc  \ \longrightarrow \   DM = %5.1f\, pc\ cm^{-3} \ \ \ \ \ \ \ \  l,\ b \ = \ %6.2f, \   %6.2f$'%(d, DM, ldeg, bdeg), fontsize=11)

    # Frame 2: spiral arms
    ax2 = fig.add_subplot(223)
    ax2.set_aspect('equal',  'box')
    for j in range(Darms['a'].size): 
        plot(armarray[0, j], armarray[1, j])
        #plot(coarse_arms[0, j], coarse_arms[1, j])
    # line of sight
    # Sun location
    plot(0, rsun, 'o', lw=0.5, ms=4, mfc='w', mec='k')
    plot(0, rsun, '.', ms=1, mfc='k', mec='k')

    # Galactic center
    plot(0, 0, '+')

    # line of sight
    plot(xvec, yvec, 'k-', lw=1)

    xlabel(r'$\rm X \ \ (kpc) $', fontsize=12)
    ylabel(r'$\rm Y \ \ (kpc) $', fontsize=12)

    # Frame 3: arm numbers
    ax3 = fig.add_subplot(224) 
    for j in range(1, Darms['a'].size+1):
       inds = np.where(Armvec==j)
       plot(s[inds], Armvec[inds], lw=2)
    xlabel(r'$\rm d \ \ (kpc) $', fontsize=12)
    ylabel(r'$\rm Arm \ number  $', fontsize=12)

    # save file in main
    #show()
    return
   
def plot_dm_ne_cn2():
    # ----------------------------------
    # Three panels with DM, n_e, and Cn2
    # ----------------------------------
    fig = figure()
    subplots_adjust(left=0.20, bottom=0.15)
    suptitle(r'$\rm Output \ from \ los\_diagnostics.py$', fontsize=12)

    # Frame 1:  DM vs s
    ax = fig.add_subplot(311)
    plot(s25, dm25)
    ylabel(r'$\rm DM \ (pc\ cm^{-3}) $', fontsize=14, labelpad=30)
    annotate(r'$\rm l = %6.1f^{\circ}  \ \ b = %5.1f$'%(ldeg, bdeg), xy=(0.025, 0.875), xycoords='axes fraction', ha='left', va='center', fontsize=10)
    annotate(r'$\rm DM_{max} =  %8.2f$'%(DMmax), xy=(0.025, 0.650), xycoords='axes fraction', ha='left', va='center', fontsize=10)
    title(r'$\rm Effective \ distances: \ \overline{d}_{n_e} = %8.2f \ \ \ \ \ \overline{d}_{n_e^2} = %8.2f \ \ \ \ \ \overline{d}_{SM} = %8.2f$'%(dbar_ne, dbar_ne2, deffsm2), fontsize=11)
    tick_params(labelbottom = False)

    # Frame 2: n_e vs s
    ax = fig.add_subplot(312)
    plot(s25, ne25, label=r'$\rm n_e$')
    ylabel(r'$\rm n_e \ (cm^{-3}) $', fontsize=14, labelpad=20)
    tick_params(labelbottom = False)

    # Frame 3: Cn2e vs s
    ax = fig.add_subplot(313)
    plot(s25, Cn2_vs_s25, label=r'$\rm C_n^2$')
    ylabel(r'$\rm  C_n^2 \ (m^{-20/3}) $', fontsize=14, labelpad=10)

    xlabel(r'$\rm Distance \ from \ solar \ system \ \ (kpc) $', fontsize=16)
    #legend(loc=0)
    #show()

    return

# Main

if __name__ == '__main__':

    try:
        parser = argparse.ArgumentParser(
            description='Plot NE20x diagnostics for designated line of sight)')

        """
        parser.add_argument('-l', '--l', default=57.51, help='Galactic longitude (deg)')
        parser.add_argument('-b', '--b', default=-0.29, help='Galactic latitude (deg)')
        parser.add_argument('-DM_D', '--DM_D', default=71, help='DM (pc cm^{-3}) or distance (kpc)')
        parser.add_argument('-ndir', '--ndir', default=1, help='ndir = 1 (DM->D), ndir = -1 (D->DM)')
        """

        parser.add_argument('l', help='Galactic longitude (deg)')
        parser.add_argument('b', help='Galactic latitude (deg)')
        parser.add_argument('DM_D', help='DM (pc cm^{-3}) or distance (kpc)')
        parser.add_argument('ndir', help='ndir = 1 (DM->D), ndir = -1 (D->DM)')
        parser.add_argument('v', default=2025,  help='v = 2025 (NE2025, default), v = 2001 (NE2001)')

        # Default values for sampling along the line of sight: less likely to need change. 
        Ncoarse_def = 50
        ds_def = 0.01
        parser.add_argument('-N', '--Ncoarse', default=Ncoarse_def, 
                help='Number of coarse values to use for smooth n_e components')
        parser.add_argument('-ds', '--ds', default=ds_def, 
                help='step size along line of sight (pc)')

        args = parser.parse_args()

        ldeg = float(args.l)
        bdeg = float(args.b)
        dmd = float(args.DM_D)
        ndir = int(args.ndir)
        Ncoarse = int(args.Ncoarse)
        ds = float(args.ds)
        v = int(args.v)

        l, b = deg2rad((ldeg, bdeg))

        if v==2001:
            # Directory for reading the f25 file written by NE20x (from call below)
            f25dir = os.getcwd()+'/output_ne2001p/'
            # Plot directory (same as f25dir here):
            plotdir = os.getcwd()+'/output_ne2001p/'
        else:
            # Directory for reading the f25 file written by NE20x (from call below)
            f25dir = os.getcwd()+'/output_ne2025p/'
            # Plot directory (same as f25dir here):
            plotdir = os.getcwd()+'/output_ne2025p/'

        print('Line of sight diagnostics for l = %6.1f   b = %6.1f   dmd = %d   ndir = %d'%(ldeg, bdeg, dmd, ndir))
        if Ncoarse != Ncoarse_def:
            print('Using Ncoarse = %d (number of samples used for smooth n_e components)'%(Ncoarse)) 
        if ds != ds_def:
            print('Using ds = %6.3f (fine sample interval for LoS)'%(ds)) 
        print('Plots are stored as PDFs in directory: \n%s'%(plotdir))

    except SystemExit:
        print('Try again with inputs:')
        print('    Use  los_diagnostics.py  l       b     DM_D ndir v  with DM_D = DM or distance for ndir > or < 0, v=2025 (NE2025) or v=2001 (NE2001)')
        print('    e.g. los_diagnostics.py 57.51 -0.29   71.02   1 2025   for B1937+21 parameters evaluated by NE2025')
        sys.exit()

    # Get output DM or d  from NE2001:
    if v==2001:
    	# set which_model.inp
        #inpdir = os.path.dirname(os.path.realpath(__file__))+'/params/'
        #with open(inpdir+"which_model.inp", "w") as file:
        #    file.write("NE2001")
        from mwprop.nemod.NE2001 import ne2001
        xx = ne2001(ldeg, bdeg, dmd, ndir, classic=False, dmd_only=False, do_analysis=True)
    if v==2025:
        # set which_model.inp
        #inpdir = os.path.dirname(os.path.realpath(__file__))+'/params/'
        #with open(inpdir+"which_model.inp", "w") as file:
        #    file.write("NE2025")
        from mwprop.nemod.NE2025 import ne2025
        xx = ne2025(ldeg, bdeg, dmd, ndir, classic=False, dmd_only=False, do_analysis=True)

    import mwprop.nemod.config_nemod as config_nemod
    from mwprop.nemod.density_components import *
    import mwprop.nemod.ne_arms as ne_arms 
    import mwprop.nemod.density_components as dc
    import mwprop.nemod.density as ne01
    import mwprop.nemod.ne_lism as lism
    import mwprop.nemod.neclumpN_fast as necf
    import mwprop.nemod.nevoidN as nev

    if ndir > 0:                # DM -> D
        DM = xx[1]['DM']
        D2001 = xx[1]['DIST']
        d = D2001
    else:
        DM2001 = xx[1]['DM']
        d = xx[1]['DIST']
        DM = DM2001

    # get spiral arm info for plotting
    Dgal01, Dgc, Dlism, Dclumps, Dvoids, Darms, Darmmap, armmap, \
           r1, th1, th1deg, coarse_arms, rarray, tharray,  armsplines, armarray, \
           tangents, normals, curvatures, eval_NE2025, eval_NE2001 = config_nemod.setup_spiral_arms(Ncoarse=Ncoarse)

    ne_arms.th1 = th1
    ne_arms.r1 = r1
    ne_arms.coarse_arms = coarse_arms
    ne_arms.Darmmap = Darmmap
    ne_arms.armmap = armmap
    ne_arms.Dgal01 = Dgal01
    
    s = np.linspace(0, d, int(d/ds))
    xvec, yvec, zvec =  s*np.sin(l), rsun-s*np.cos(l), s*np.sin(b)  

    nea_vec = zeros(np.size(xvec))
    ne1_vec = zeros(np.size(xvec))
    ne2_vec = zeros(np.size(xvec))
    negc_vec = zeros(np.size(xvec))
    Fvec = zeros(np.size(xvec))
    Armvec = zeros(np.size(xvec), dtype=int)
 
    ne_lism_vec = zeros(np.size(xvec))
    F_lism_vec = zeros(np.size(xvec))
    wlism_vec = zeros(np.size(xvec))

    for n, x in enumerate(xvec):

        # Get individual electron density contributions
        nea_vec[n], Fvec[n], Armvec[n] = \
            ne_arms.ne_arms_ne2001p(x ,yvec[n], zvec[n], Ncoarse=Ncoarse, verbose=False)
        ne1_vec[n] = dc.ne_outer(x, yvec[n], zvec[n])[0] 
        ne2_vec[n] = dc.ne_inner(x, yvec[n], zvec[n])[0] 
        negc_vec[n] = dc.ne_gc(x, yvec[n], zvec[n])[0] 

        ne_LISM, FLISM, wLISM, wLDR, wLHB, wLSB, wLOOPI = lism.ne_lism(xvec[n], yvec[n], zvec[n])
        ne_lism_vec[n] = ne_LISM
        F_lism_vec[n] = FLISM
        wlism_vec[n] = wLISM

    # Get total n_e from file f25 text file written by analysis_dmd_dm_and_sm() in dmdsm.py
    if ndir < 0:
        f25file = 'f25_d2dm_ne_dsm_vs_s.txt'
    else:
        f25file = 'f25_dm2d_ne_dsm_vs_s.txt'
        
    # Read f25 output file written by NE20x into standard directory:
    ldeg, bdeg = np.loadtxt(f25dir+f25file, unpack=True, usecols = (0, 1), skiprows=1, max_rows=1)
    s25, ne25, Cn2_vs_s25, dm25, nea25 = np.loadtxt(f25dir + f25file, unpack=True, usecols = (0, 4, 5, 10, 11), skiprows=3)

    DMmax = dm25.max()

    # Effective distances based on LoS integrals weighted by various functions of n_e:
    dbar_ne = np.trapz(s25*ne25, s25) / np.trapz(ne25, s25)
    dbar_ne2 = np.trapz(s25*ne25**2, s25) / np.trapz(ne25**2, s25)
    deffsm2 =  np.trapz(s25*Cn2_vs_s25, s25) / np.trapz(Cn2_vs_s25, s25)

    # Cumulative DM and SM: not currently used, so commented out; may use later so keep.
    #DMvec = 1000.*np.array([np.trapz(ne25[:j], s25[:j]) for j in range(np.size(s25))])
    #SMvec = np.array([np.trapz(Cn2_vs_s25[:j], s25[:j]) for j in range(np.size(s25))])


    plot_ne_arms()
    plotfile = 'ne_vs_d_arms_' + '%6.1f'%(ldeg) + '_' + '%5.1f'%(bdeg) + '_' + str(int(DMmax)) + '_' +  basename + '.pdf'
    plotfile = plotfile.replace(' ', '')
    savefig(plotdir+plotfile)
    show()

    plot_dm_ne_cn2()
    plotfile = 'dm_ne_cn2_l_b_dmmax_' + '%6.1f'%(ldeg) + '_' + '%5.1f'%(bdeg) + '_' + str(int(DMmax)) + '_' +  basename + '.pdf'
    plotfile = plotfile.replace(' ', '')
    savefig(plotdir+plotfile)
    show()
    
    input('hit return')
    close('all')



# mwprop v2.0 Jan 2026

"""
ne_lism.py

Local interstellar medium functions

Replacements for Fortran functions in neLISM.NE2001.f:
    ne_LISM(x,y,z,FLISM,wLISM)      ! total local ISM 
    neLDRQ1(x,y,z,FLDRQ1r,wLDRQ1)   ! Low Density Region in Q1
    neLSB(x,y,z,FLSBr,wLSB)         ! Local Super Bubble
    neLHB(x,y,z,FLHBr,wLHB)         ! Local Hot Bubble [not used]
    neLHB2(x,y,z,FLHBr,wLHB)        ! Local Hot Bubble
    neLOOPI(x,y,z,FLOOPI,wLOOPI)    ! Loop I

Python version JMC 2020 Jan 27

Changes:
2022 Jan 02: 
    1. added if statement to skip calculations for x,y values
       outside of LISM
    2. put precalculations of some parameters into config_ne2001p.py
    3. annotations added

Comments from Fortran code neLISM.NE2001.f
c routines to calculate the electron density for the
c Local Interstellar Medium
c
c JMC 26 August-11 Sep. 2000
c     25 October 2001: modified to change weighting scheme
c                      so that the ranking is LHB: LSB: LDR
c                      (LHB overrides LSB and LDR; LSB overrides LDR)
c     16 November 2001: 
c           added Loop I component with weighting scheme
c               LHB:LOOPI:LSB:LDR
c               LHB   overides everything,
c           LOOPI overrides LSB and LDR
c           LSB   overrides LDR
c           LISM  overrides general Galaxy
c     20 November 2001: The LOOPI component is truncated below z=0
c
c after discussions with Shami Chatterjee
c the sizes, locations and densities of the LISM components
c are based largely on work by Toscano et al. 1999 and Bhat et al. 1999

Note mwprop needs to be in PYTHONPATH

"""


from mwprop.nemod.config_nemod import *

script_path = os.path.dirname(os.path.realpath(__file__))


# -----------------------------------------------------------------------------

def ne_lism(x,y,z):
    """
    Input variables:
        x,y,z = NE2001 Galactocentric coordinates (kpc)

    Output:
        ne_LISM = electron density
        FLISM = fluctuation parameter
        wLISM = weight for composite LISM 
        wLDR = weight for LDR region
        wLHB = weight for local hot bubble
        wLSB = weight 
        wLOOPI = weight for radio loop I

    Based on original Fortran routine.

    Changes:
        2022 Jan 02:  put in test for y out of range of LISM; return zeros if so
                      note: y_lism_min, y_lism_max computed in config_ne2001p.py 
    """

    # Test whether input y is enclosed by LISM; if not set values to zero and return

    if y > y_lism_max or y < y_lism_min:
        return 0, 0, 0, 0, 0, 0, 0 

    neldrq1xyz, FLDRQ1r, wLDR = neLDRQ1(x,y,z)    # low density region in Q1
    nelsbxyz, FLSBr, wLSB = neLSB(x,y,z)          # Local Super Bubble
    nelhbxyz, FLHBr, wLHB = neLHB2(x,y,z)         # Local Hot Bubble
    neloopIxyz, FLOOPIr, wLOOPI = neLOOPI(x,y,z)  # Loop I

    """
    weight the terms so that the LHB term overrides the other
    terms (we want the density to be low in the LHB, lower than
    in the other terms.
    """

    ne_LISM =   ((1-wLHB)   *
                 (
                   (1-wLOOPI) * (wLSB*nelsbxyz + (1-wLSB)*neldrq1xyz)
               +     wLOOPI * neloopIxyz
                 )
               +     wLHB  * nelhbxyz)

    FLISM = ((1-wLHB) *
                (
                   (1-wLOOPI) * (wLSB*FLSBr + (1-wLSB)*FLDRQ1r)
               +     wLOOPI * FLOOPIr
                )
               +     wLHB  * FLHBr)

    """
    return the maximum weight of any of the terms for
    combining with additional terms external to this routine.
    """

    wLISM = max(wLOOPI, max(wLDR, max(wLSB, wLHB)))

    return ne_LISM, FLISM, wLISM, wLDR, wLHB, wLSB, wLOOPI
    return

# -----------------------------------------------------------------------------

def neLDRQ1(x,y,z):    # Low Density Region in Q1
    """
    Low Density Region in Q1
    input:
      x,y,z = coordinates w.r.t. Galaxy as in TC93, CL00

    output:
      neLDRQ1 = electron density in local hot bubble that
              is modeled as an ellipsoidal trough.
      FLDRQ1 = fluctuation parameter
      wLDRQ1  = weight of LDRQ1 component used to combine
          with other components of electron density.
          wLDRQ1 =  1  at and inside the annular ridge
                 <  1  outside the annular ridge
                 -> 0  far outside the annular ridge
      e.g. total electron density would be evaluated as
               ne = (1-wLDRQ1)*ne_other + neLDRQ1
      note thetaldr is converted to radians in config_ne2001p.py

    These calculations now done in config_ne2001p.py
    aa=aldr
    bb=bldr
    cc=cldr
    theta=thetaldr
    netrough =neldr0
    Ftrough=Fldr

    # can redef s,c as sldr, cldr and do these calcs outside function in setup
    s = np.sin(theta)
    c = np.cos(theta)

    ap = (c/aa)**2 + (s/bb)**2
    bp = (s/aa)**2 + (c/bb)**2
    cp = 1./cc**2
    dp =  2.*c*s*(1./aa**2 - 1./bb**2)
    sldr = np.sin(thetaldr)
    cldr = np.cos(thetaldr)

    apldr = (cldr/aldr)**2 + (sldr/bldr)**2
    bpldr = (sldr/aldr)**2 + (cldr/bldr)**2
    cpldr = 1./cldr**2
    dpldr =  2.*cldr*sldr*(1./aldr**2 - 1./bldr**2)
    """

    neLDRQ1 = 0.
    wLDRQ1 = 0
    FLDRQ1r = 0.
    q = ((x-xldr)**2*apldr
          + (y-yldr)**2*bpldr
          + (z-zldr)**2*cpldr
          + (x-xldr)*(y-yldr)*dpldr)
    if q <= 1:                     # inside
        neLDRQ1 = neldr0
        FLDRQ1r = Fldr
        wLDRQ1 = 1

    return neLDRQ1, FLDRQ1r, wLDRQ1

# -----------------------------------------------------------------------------

def neLSB(x,y,z):                       # Local Super Bubble
    """
    Local Super Bubble

    input:
      x,y,z = coordinates w.r.t. Galaxy as in TC93, CL00
    output:
      neLSB = electron density in local hot bubble that
              is modeled as an ellisoidal trough.
      FLSB = fluctuation parameter
      wLSB  = weight of LSB component used to combine
          with other components of electron density.
          wLSB =  1  at and inside the annular ridge
               <  1  outside the annular ridge
                   -> 0  far outside the annular ridge
      e.g. total electron density would be evaluated as
      ne = (1-wLSB)*ne_other + neLSB

    These now calculated in input_ne2001p.py

    aa=alsb
    bb=blsb
    cc=clsb
    theta=thetalsb
    netrough=nelsb0
    Ftrough=Flsb

    s = np.sin(theta)
    c = np.cos(theta)
    ap = (c/aa)**2 + (s/bb)**2
    bp = (s/aa)**2 + (c/bb)**2
    cp = 1./cc**2
    dp =  2.*c*s*(1./aa**2 - 1./bb**2)
    """

    neLSB = 0.
    wLSB = 0
    FLSBr = 0.
    q = ((x-xlsb)**2*aplsb
          + (y-ylsb)**2*bplsb
          + (z-zlsb)**2*cplsb
          + (x-xlsb)*(y-ylsb)*dplsb)
    if q <= 1:                      # inside region
        neLSB = nelsb0
        FLSBr = Flsb
        wLSB = 1

    return neLSB, FLSBr, wLSB

# -----------------------------------------------------------------------------

def neLHB(x,y,z):                 # Local Hot Bubble
    """
    Local Hot Bubble
    input:
      x,y,z = coordinates w.r.t. Galaxy as in TC93, CL00
    output:
      neLHB = electron density in local hot bubble that
              is modeled as an ellisoidal trough.
      FLHB = fluctuation parameter
      wLHB  = weight of LBH component used to combine
          with other components of electron density.
          wLBH =  1  at and inside the annular ridge
               <  1  outside the annular ridge
                   -> 0  far outside the annular ridge
      e.g. total electron density would be evaluated as
               ne = (1-wLHB)*ne_other + neLHB

    These are now calculated in input_ne2001p.py
    aa=alhb
    bb=blhb
    cc=clhb
    theta=thetalhb
    netrough=nelhb0
    Ftrough=Flhb

    s = np.sin(theta)
    c = np.cos(theta)
    ap = (c/aa)**2 + (s/bb)**2
    bp = (s/aa)**2 + (c/bb)**2
    cp = 1./cc**2
    dp = 2.*c*s*(1./aa**2 - 1./bb**2)
    """

    neLHB = 0.
    wLHB = 0
    FLHBr = 0.
    q = ((x-xlhb)**2*aplhb
      + (y-ylhb)**2*bplhb
      + (z-zlhb)**2*cplhb
      + (x-xlhb)*(y-ylhb)*dplhb)
    if q <= 1:                           # inside
        neLHB = nelhb0
        FLHBr = Flhb
        wLHB = 1

    return neLHB, FLHBr, wLHB

# -----------------------------------------------------------------------------

def neLHB2(x,y,z):                        # Local Hot Bubble
    """
    LHB modeled as a cylinder
    the cylinder slants in the y direction vs. z as described by parameter yzslope
    the cylinder cross-sectional size in the 'a' direction (major axis)
          varies with z, tending to zero at its smallest z point.
        implicit none
        real x,y,z,FLHBr
        integer wLHB

    input:
      x,y,z = coordinates w.r.t. Galaxy as in TC93, CL00

    output:
      neLHB2 = electron density in local hot bubble that
              is modeled as an ellisoidal trough.
      FLHB = fluctuation parameter
      wLHB  = weight of LBH component used to combine
          with other components of electron density.
          wLHB =  1  at and inside the annular ridge
               <  1  outside the annular ridge
                   -> 0  far outside the annular ridge
      e.g. total electron density would be evaluated as
               ne = (1-wLHB)*ne_other + neLHB2
    aa=alhb
    bb=blhb
    cc=clhb
    netrough=nelhb0
    Ftrough=Flhb
    theta = thetalhb 
    """

    neLHB2 = 0.
    wLHB = 0
    FLHBr = 0.

    yzslope = np.tan(thetalhb)
    yaxis = ylhb + yzslope*z

    """
    cylinder has cross sectional area = constant for z > 0
    area -> 0 for z < 0 by letting aa->0 linearly for z < 0:
    (0.001 = 1 pc is to avoid divide by zero)
    """

    if z <= 0 and z >= zlhb-clhb:
      aa = 0.001 + (alhb-0.001)*(1. - (1./(zlhb-clhb))*z)
    else:
      aa = alhb
    qxy =  ((x-xlhb)/aa)**2 + ((y-yaxis)/blhb)**2
    qz =  abs(z-zlhb)/clhb
    if qxy <= 1 and qz <= 1:            # inside
        neLHB2 = nelhb0
        FLHBr = Flhb
        wLHB = 1

    return neLHB2, FLHBr, wLHB

# -----------------------------------------------------------------------------

def neLOOPI(x,y,z):                      # Loop I
    """
    Loop I

    Component is a spheroid truncated for z<0.

    input:
      x,y,z = coordinates w.r.t. Galaxy as in TC93, CL00

    output:
      neLOOPI = electron density in LOOP I that
              is modeled as an ellisoidal trough
          with an enhanced shell
      FLOOPI = fluctuation parameter
      wLOOPI  = weight of LOOP I component used to combine
          with other components of electron density.
          wLOOPI =  1  at and inside the annular ridge
                 <  1  outside the annular ridge
    """
    a1 = rlpI
    a2 = rlpI+drlpI

    if z < 0:
        neLOOPI = 0.
        FLOOPI = 0.
        wLOOPI = 0
        return neLOOPI, FLOOPI, wLOOPI

    r = sqrt( (x-xlpI)**2 + (y-ylpI)**2 + (z-zlpI)**2)
    if r > a2:                  # outside Loop I
        neLOOPI = 0.
        FLOOPI = 0.
        wLOOPI = 0
    elif r <= a1:               # inside volume
        neLOOPI= nelpI
        FLOOPI = FlpI
        wLOOPI = 1
    else:                       # inside boundary shell
        neLOOPI= dnelpI
        FLOOPI = dFlpI
        wLOOPI = 1

    return neLOOPI, FLOOPI, wLOOPI

# -----------------------------------------------------------------------------



# mwprop.ne2001p v1.0 Jan 2024

from __future__ import print_function
from numpy import sqrt
from math import pi
from scipy import constants

"""
Gives constants needed for interstellar wave propagation and other applications

Units: cgs
"""

# Solar mass
# from astropy.constants.M_sun.cgs.value (scipy.constants doesn't have!)
msun = 1.9884754153381438e+33

# fundamental
eV = constants.eV / constants.erg
c = constants.c *  100.         # cm
r_e = constants.physical_constants['classical electron radius'][0]*100. # cm
m_e = constants.constants.m_e*1000
e_e = sqrt(m_e * c**2 * r_e)
kb = constants.k / constants.erg
G = constants.G / constants.gram 

h = constants.h / constants.erg
hbar = constants.hbar / constants.erg

# lengths 

cm_in_m = 100.
km = 1.e5                       # cm
au = constants.au * 100.        # cm
pc = constants.parsec * 100.    # cm
kpc = 1000.*pc
kpc = 1.e3*pc
Mpc = 1.e6*pc
Gpc = 1.e9*pc

# angular

mas = pi / (180. * 3600.*1000.)
ns = 1.e-9
all_sky_sr = 4.*pi
all_sky_degsq = (180./pi)**2 * all_sky_sr

# frequencies

MHz = 1.e6
GHz = 1.e9

# temporal
ns = 1.e-9       # sec
mus = 1.e-6      # sec
ms = 1.e-3       # sec


# Dispersion constant
KDM = (c * r_e * pc) / (2.*pi*GHz**2)                # for TOAs in seconds

# Faraday rotation measure constant

microGauss = 1.e-6

# RMconstant gives standard rad m^{-2}
# for B in microgauss, pathlength in pc, electron density in cm^{-3}

RMconstant = ((e_e**3) / (2.*pi*m_e**2 * c**4)) * pc * cm_in_m**2 * microGauss
RMconstant_cgs = ((e_e**3) / (2.*pi*m_e**2 * c**4)) * pc * microGauss

# Scattering measure kpc m^{-20/3} to cgs
SMunit = kpc*10.**(-40./3.)             # kpc m^-20/3 to cgs
SM1 = 1.                                # kpc m^{-20/3}
SM_cgs = SM1 * kpc * cm_in_m**(-20./3.) # same as SMunit (a check)


# Flux density

Jy = 1.e-23     # cgs: erg/s/cm^2/Hz
Jansky = Jy

# Brightness temperature for 1 Jy pulse of 1 ms at 1 kpc at 1 GHz
Tb111 = (Jy * kpc**2) / (2.*kb*(GHz*ms)**2)

# plasma frequency in Hz for unit electron density
ne1 = 1.                                                # 1/cc
nup = sqrt((4.*pi * ne1 * e_e**2) / m_e) / (2.*pi)      # Hz

# cyclotron frequency in a 1 Gauss field
B1=1.
nuc = (e_e * B1) / (m_e * c) / (2.*pi)                  # Hz
